#include "stdtypes.h"

void    quantize();
void    iquantize();
long    divide();
long    range_check();
void    zig_zag();
void    zag_zig();
long    sign();

/* sequence of quantization */  	
static int zagTable[64] = { 
		 0,  1,  5,  6, 14, 15, 27, 28,
		 2,  4,  7, 13, 16, 26, 29, 42,
		 3,  8, 12, 17, 25, 30, 41, 43,
		 9, 11, 18, 24, 31, 40, 44, 53,
	        10, 19, 23, 32, 39, 45, 52, 54,
	        20, 22, 33, 38, 46, 51, 55, 60,
	        21, 34, 37, 47, 50, 56, 59, 61,
	        35, 36, 48, 49, 57, 58, 62, 63  
};

/* quantization table for a luminance block */
static int lu_table[64] = { 
		  16,  11,  12,  14,  12,  10,  16,  14,
		  13,  14,  18,  17,  16,  19,  24,  40,
		  26,  24,  22,  22,  24,  49,  35,  37,
		  29,  40,  58,  51,  61,  60,  57,  51,
		  56,  55,  64,  72,  92,  78,  64,  68,
		  87,  69,  55,  56,  80, 109,  81,  87,
		  95,  98, 103, 104, 103,  62,  77, 113,
		 121, 112, 100, 120,  92, 101, 103,  99
};

/* quantization table for a chrominance block */
static int cr_table[64] = { 
		  17,  18,  18,  24,  21,  24,  47,  26,
		  26,  47,  99,  66,  56,  66,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99,
		  99,  99,  99,  99,  99,  99,  99,  99
};

/****************************************************************************
 Objectives : adjust a value in a certain range
 Input      : one long integer, upper bound, under bound
 Output     : adjusted value
 Called By  : quantize(), iquantize()
 Calls      : none
****************************************************************************/
long range_check(data, lower_bound, upper_bound) 
long   data, lower_bound, upper_bound;
{
	if (data < lower_bound) 
		return lower_bound;
	else if( data > upper_bound) 
		return upper_bound;
	else	
		return data;
};

/****************************************************************************
 Objectives : notify if a value is negative, positive, or zero
 Input      : one long integer
 Output     : 0, -1, or 1
 Called By  : 1quantize()
 Calls      : none
****************************************************************************/
long sign(data)
long  data;
{
	if(!data) return 0;
	if(data<0) return -1;
	return 1;
}

/****************************************************************************
 Objectives : division
 Input      : two long integers
 Output     : the result of division
 Called By  : quantize(), 1quantize()
 Calls      : none
****************************************************************************/
long divide(num, denum) 
long  num, denum;
{
	double data;

	data = (double)num / (double)denum;
	if ( data > 0 )
		return (long)(data+0.5);
	else return (long)(data-0.5);
};

/****************************************************************************
 Objectives : quantizing blocks in a zig-zag sequence
 Input      : a DCT block, frame type, quantization factor
 Output     : quantized block
 Called By  : quantize()
 Calls      : divide(), range_check()
****************************************************************************/
void zig_zag(patch, block, quant, frame_type) 
long  *patch;
BLOCK *block;
int   quant, frame_type;
{
	long 		data, buffer[64];
	int		*weight;  /* pointer to quantization table */
 	int 		i;        /* index */
	
	weight = lu_table;
	if(frame_type == UFRAME || frame_type == VFRAME) 
		weight = cr_table;

	/* DC value */
	buffer[0] = range_check(divide(patch[0], weight[0]), -128, 127); 
	block->element[0][0] = buffer[0];

	/* AC values */
	for(i=1; i<64; i++) {  
		data = divide(patch[i], weight[i]);
		data = (double)data / ((quant-1)/5*0.1 + 1.0);
		buffer[zagTable[i]] = range_check(data, -128, 127);
	}
	for(i=1; i<64; i++)  
		block->element[i/8][i%8] = buffer[i];
};

/****************************************************************************
 Objectives : inverse quantizing blocks in a zig-zag sequence
 Input      : a quantized block, frame type, quantization factor
 Output     : quantized block
 Called By  : iquantize()
 Calls      : divide(), range_check(), sign()
****************************************************************************/
void zag_zig(buffer, block, quant, frame_type) 
long  *buffer;
long block[][8];
int   quant, frame_type;
{
	long 		data, patch[64];
	int		*weight, i;
	
	weight = lu_table;
	if(frame_type == UFRAME || frame_type == VFRAME) 
		weight = cr_table;

	/* DC value */
	patch[0] = buffer[0] * weight[0]; 
	block[0][0] = patch[0];

	/* AC values */
	for(i=1; i<64; i++) {  
		data = buffer[zagTable[i]];
		data = (data * ((quant-1)/5*0.1 + 1.0) * weight[i]);
		if ( ( data & 1 ) == 0 )
			data -= sign( data );
		patch[i] = range_check( data, -1024, 1023 );
		block[i/8][i%8] = patch[i];
	}
};

/****************************************************************************
 Objectives : quantizing blocks
 Input      : a DCT block, frame type, quantization factor
 Output     : quantized block
 Called By  : mcu_coding()
 Calls      : zig_zag()
****************************************************************************/
void quantize(patch, mcu_block, quant, frame_type)
long patch[][8]; 
BLOCK  *mcu_block;
int   quant, frame_type;
{
	long *ptr;
	ptr = &(patch[0][0]);
	zig_zag(ptr, mcu_block, quant, frame_type);
};

/****************************************************************************
 Objectives : inverse quantizing blocks
 Input      : a quantized block, frame type, quantization factor
 Output     : inverse quantized block
 Called By  : mcu_coding()
 Calls      : zag_zig()
****************************************************************************/
void iquantize(patch, block, quant, frame_type)
BLOCK *patch;
long block[][8];
int   quant, frame_type;
{
	long *ptr;

	ptr = &(patch->element[0][0]);
	zag_zig(ptr, block, quant, frame_type);
};

